%% Implementation of the single shooting scheme of optimal control problems
%close all
clear variables
clc

%% Create NLP
import casadi.*

N   = 30;       % Control discretization
T   = 3.0;      % End time
h   = T/N;      % Discretization step
nx  = 1;            
nu  = 1;

% Tuning
Q   = 1;
R   = 1;

% Declare variables (use scalar graph)
u  = SX.sym('u',nu);    % control
x  = SX.sym('x',nx);    % states

% YOUR CODE HERE: System dynamics
% xdot = ...;

f = Function('f', {x,u},{xdot});

% RK4
U   = MX.sym('U');
X0  = MX.sym('X0',nx);
X   = X0;

% YOUR CODE HERE: implement the RK4 integrator
RK4 = Function('RK4',{X0,U},{X});

% Formulate NLP
u = SX.sym('u',N);

% Get an expression for the cost and state at end
x_0 = 0.05;
X = x_0;

% Objective function
J=0;
for i = 1:N
    % YOUR CODE HERE: build the cost of the NLP
    % 1 - get x_next using RK4
    % 2 - J = J + ...
end

% Terminal constraints: x_0(T)=x_1(T)=0
g = X;

% Allocate an NLP solver
nlp = struct('x', u, 'f', J, 'g', g);

% Create IPOPT solver object
options = struct;
options.hessian_approximation = 'limited-memory';
solver = nlpsol('solver', 'ipopt', nlp);

arg = struct;
% YOUR CODE HERE: upper and lower bounds on x and g(x)
% arg.x0  =  0.;        % solution guess
% arg.lbx =  ...;       % lower bound on x
% arg.ubx =  ...;       % upper bound on x
% arg.lbg =  ...;       % lower bound on g
% arg.ubg =  ...;       % upper bound on g

%% Solve the NLP
res = solver(arg);

f_opt       = full(res.f);
u_opt       = full(res.x);

% Compute state trajectory integrating the dynamics
x_opt       = zeros(nx,N+1);
x_opt(1)    = x_0;

for i=2:N+1
    out         = RK4({x_opt(i-1),u_opt(i-1)});
    x_opt(i)    = full(out{1});
end

%% Plot results

figure()
subplot(2,1,1)
plot((0:N)*T/N, x_opt)
xlabel('time - t')
ylabel('state - x')
grid on
subplot(2,1,2)
stairs((0:N-1)*T/N, u_opt,'r')
xlabel('time - t')
ylabel('input - u')
grid on

